import azure.cosmos.documents as documents
import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.errors as errors

import samples.Shared.config as cfg

# ----------------------------------------------------------------------------------------------------------
# Prerequistes - 
# 
# 1. An Azure Cosmos account - 
#    https://azure.microsoft.com/en-us/documentation/articles/documentdb-create-account/
#
# 2. Microsoft Azure Cosmos PyPi package - 
#    https://pypi.python.org/pypi/azure-cosmos/
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates the basic CRUD operations on a Collection resource for Azure Cosmos
# 
# 1. Query for Collection
#  
# 2. Create Collection
#    2.1 - Basic Create
#    2.2 - Create collection with custom IndexPolicy
#    2.3 - Create collection with offer throughput set
#    2.4 - Create collection with unique key
#    2.5 - Create Collection with partition key
#    2.6 - Create Collection with partition key V2
#
# 3. Manage Collection Offer Throughput
#    3.1 - Get Collection performance tier
#    3.2 - Change performance tier
#
# 4. Get a Collection by its Id property
#
# 5. List all Collection resources in a Database
#
# 6. Delete Collection
# ----------------------------------------------------------------------------------------------------------
# Note - 
# 
# Running this sample will create (and delete) multiple DocumentContainers on your account. 
# Each time a DocumentContainer is created the account will be billed for 1 hour of usage based on
# the performance tier of that account. 
# ----------------------------------------------------------------------------------------------------------

HOST = cfg.settings['host']
MASTER_KEY = cfg.settings['master_key']
DATABASE_ID = cfg.settings['database_id']
COLLECTION_ID = cfg.settings['collection_id']

database_link = 'dbs/' + DATABASE_ID

class IDisposable(cosmos_client.CosmosClient):
    """ A context manager to automatically close an object with a close method
    in a with statement. """

    def __init__(self, obj):
        self.obj = obj

    def __enter__(self):
        return self.obj # bound to target

    def __exit__(self, exception_type, exception_val, trace):
        # extra cleanup in here
        self.obj = None

class CollectionManagement:
    @staticmethod
    def find_Container(client, id):
        print('1. Query for Collection')

        collections = list(client.QueryContainers(
            database_link,
            {
                "query": "SELECT * FROM r WHERE r.id=@id",
                "parameters": [
                    { "name":"@id", "value": id }
                ]
            }
        ))

        if len(collections) > 0:
            print('Collection with id \'{0}\' was found'.format(id))
        else:
            print('No collection with id \'{0}\' was found'. format(id))
        
    @staticmethod
    def create_Container(client, id):
        """ Execute the most basic Create of collection. 
        This will create a collection with 400 RUs throughput and default indexing policy """

        print("\n2.1 Create Collection - Basic")
        
        try:
            client.CreateContainer(database_link, {"id": id})
            print('Collection with id \'{0}\' created'.format(id))

        except errors.HTTPFailure as e:
            if e.status_code == 409:
               print('A collection with id \'{0}\' already exists'.format(id))
            else: 
                raise

        print("\n2.2 Create Collection - With custom index policy")
        
        try:
            coll = {
                "id": "collection_custom_index_policy",
                "indexingPolicy": {
                    "indexingMode": "lazy",
                    "automatic": False
                }
            }

            collection = client.CreateContainer(database_link, coll)
            print('Collection with id \'{0}\' created'.format(collection['id']))
            print('IndexPolicy Mode - \'{0}\''.format(collection['indexingPolicy']['indexingMode']))
            print('IndexPolicy Automatic - \'{0}\''.format(collection['indexingPolicy']['automatic']))
            
        except errors.CosmosError as e:
            if e.status_code == 409:
               print('A collection with id \'{0}\' already exists'.format(collection['id']))
            else: 
                raise

        print("\n2.3 Create Collection - With custom offer throughput")

        try:
            coll = {"id": "collection_custom_throughput"}
            collection_options = { 'offerThroughput': 400 }
            collection = client.CreateContainer(database_link, coll, collection_options )
            print('Collection with id \'{0}\' created'.format(collection['id']))
            
        except errors.HTTPFailure as e:
            if e.status_code == 409:
               print('A collection with id \'{0}\' already exists'.format(collection['id']))
            else: 
                raise

        print("\n2.4 Create Collection - With Unique keys")

        try:
            coll = {"id": "collection_unique_keys", 'uniqueKeyPolicy': {'uniqueKeys': [{'paths': ['/field1/field2', '/field3']}]}}
            collection_options = { 'offerThroughput': 400 }
            collection = client.CreateContainer(database_link, coll, collection_options )
            unique_key_paths = collection['uniqueKeyPolicy']['uniqueKeys'][0]['paths']
            print('Collection with id \'{0}\' created'.format(collection['id']))
            print('Unique Key Paths - \'{0}\', \'{1}\''.format(unique_key_paths[0], unique_key_paths[1]))
            
        except errors.HTTPFailure as e:
            if e.status_code == 409:
               print('A collection with id \'{0}\' already exists'.format(collection['id']))
            else: 
                raise

        print("\n2.5 Create Collection - With Partition key")
        
        try:
            coll = {
                "id": "collection_partition_key",
                "partitionKey": {
                    "paths": [
                      "/field1"
                    ],
                    "kind": "Hash"
                }
            }

            collection = client.CreateContainer(database_link, coll)
            print('Collection with id \'{0}\' created'.format(collection['id']))
            print('Partition Key - \'{0}\''.format(collection['partitionKey']))
            
        except errors.CosmosError as e:
            if e.status_code == 409:
               print('A collection with id \'{0}\' already exists'.format(collection['id']))
            else: 
                raise

        print("\n2.6 Create Collection - With Partition key V2")

        try:
            coll = {
                "id": "collection_partition_key_v2",
                "partitionKey": {
                    "paths": [
                        "/field1"
                    ],
                    "kind": "Hash",
                    "version": 2
                }
            }

            collection = client.CreateContainer(database_link, coll)
            print('Collection with id \'{0}\' created'.format(collection['id']))
            print('Partition Key - \'{0}\''.format(collection['partitionKey']))

        except errors.CosmosError as e:
            if e.status_code == 409:
                print('A collection with id \'{0}\' already exists'.format(collection['id']))
            else:
                raise

    @staticmethod
    def manage_offer_throughput(client, id):
        print("\n3.1 Get Collection Performance tier")
        
        #A Collection's Offer Throughput determines the performance throughput of a collection. 
        #A Collection is loosely coupled to Offer through the Offer's offerResourceId
        #Offer.offerResourceId == Collection._rid
        #Offer.resource == Collection._self
        
        try:
            # read the collection, so we can get its _self
            collection_link = database_link + '/colls/{0}'.format(id)
            collection = client.ReadContainer(collection_link)

            # now use its _self to query for Offers
            offer = list(client.QueryOffers('SELECT * FROM c WHERE c.resource = \'{0}\''.format(collection['_self'])))[0]
            
            print('Found Offer \'{0}\' for Collection \'{1}\' and its throughput is \'{2}\''.format(offer['id'], collection['_self'], offer['content']['offerThroughput']))

        except errors.HTTPFailure as e:
            if e.status_code == 404:
                print('A collection with id \'{0}\' does not exist'.format(id))
            else: 
                raise

        print("\n3.2 Change Offer Throughput of Collection")
                           
        #The Offer Throughput of a collection controls the throughput allocated to the Collection
        #To increase (or decrease) the throughput of any Collection you need to adjust the Offer.content.offerThroughput
        #of the Offer record linked to the Collection
        
        #The following code shows how you can change Collection's throughput
        offer['content']['offerThroughput'] += 100
        offer = client.ReplaceOffer(offer['_self'], offer)

        print('Replaced Offer. Offer Throughput is now \'{0}\''.format(offer['content']['offerThroughput']))
                                
    @staticmethod
    def read_Container(client, id):
        print("\n4. Get a Collection by id")

        try:
            # All Azure Cosmos resources are addressable via a link
            # This link is constructed from a combination of resource hierachy and 
            # the resource id. 
            # Eg. The link for collection with an id of Bar in database Foo would be dbs/Foo/colls/Bar
            collection_link = database_link + '/colls/{0}'.format(id)

            collection = client.ReadContainer(collection_link)
            print('Collection with id \'{0}\' was found, it\'s _self is {1}'.format(collection['id'], collection['_self']))

        except errors.HTTPFailure as e:
            if e.status_code == 404:
               print('A collection with id \'{0}\' does not exist'.format(id))
            else: 
                raise
    
    @staticmethod
    def list_Containers(client):
        print("\n5. List all Collection in a Database")
        
        print('Collections:')
        
        collections = list(client.ReadContainers(database_link))
        
        if not collections:
            return

        for collection in collections:
            print(collection['id'])          
        
    @staticmethod
    def delete_Container(client, id):
        print("\n6. Delete Collection")
        
        try:
           collection_link = database_link + '/colls/{0}'.format(id)
           client.DeleteContainer(collection_link)

           print('Collection with id \'{0}\' was deleted'.format(id))

        except errors.HTTPFailure as e:
            if e.status_code == 404:
               print('A collection with id \'{0}\' does not exist'.format(id))
            else: 
                raise

def run_sample():

    with IDisposable(cosmos_client.CosmosClient(HOST, {'masterKey': MASTER_KEY} )) as client:
        try:
            # setup database for this sample
            try:
                client.CreateDatabase({"id": DATABASE_ID})
            
            except errors.HTTPFailure as e:
                if e.status_code == 409:
                   pass
                else: 
                    raise
            
            # query for a collection            
            CollectionManagement.find_Container(client, COLLECTION_ID)
            
            # create a collection
            CollectionManagement.create_Container(client, COLLECTION_ID)
            
            # get & change Offer Throughput of collection
            CollectionManagement.manage_offer_throughput(client, COLLECTION_ID)

            # get a collection using its id
            CollectionManagement.read_Container(client, COLLECTION_ID)

            # list all collection on an account
            CollectionManagement.list_Containers(client)

            # delete collection by id
            CollectionManagement.delete_Container(client, COLLECTION_ID)

            # cleanup database after sample
            try:
                client.DeleteDatabase(database_link)
            
            except errors.CosmosError as e:
                if e.status_code == 404:
                   pass
                else: 
                    raise errors.HTTPFailure(e.status_code)

        except errors.HTTPFailure as e:
            print('\nrun_sample has caught an error. {0}'.format(e))
        
        finally:
            print("\nrun_sample done")

if __name__ == '__main__':
    try:
        run_sample()

    except Exception as e:
        print("Top level Error: args:{0}, message:{1}".format(e.args,e))



#The MIT License (MIT)
#Copyright (c) 2014 Microsoft Corporation

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

"""Internal class for query execution endpoint component implementation in the Azure Cosmos database service.
"""
import numbers

from azure.cosmos.execution_context.aggregators import _AverageAggregator, _CountAggregator, _MaxAggregator, \
    _MinAggregator, _SumAggregator


class _QueryExecutionEndpointComponent(object):
    def __init__(self, execution_context):
        self._execution_context = execution_context

    def __iter__(self):
        return self

    def next(self):
        return next(self._execution_context)

    def __next__(self):
        # supports python 3 iterator
        return self.next()

class _QueryExecutionOrderByEndpointComponent(_QueryExecutionEndpointComponent):
    """Represents an endpoint in handling an order by query.

    For each processed orderby result it returns 'payload' item of the result
    """
    def __init__(self, execution_context):
        super(_QueryExecutionOrderByEndpointComponent, self).__init__(execution_context)

    def next(self):
        return next(self._execution_context)['payload']

class _QueryExecutionTopEndpointComponent(_QueryExecutionEndpointComponent):
    """Represents an endpoint in handling top query.

    It only returns as many results as top arg specified.
    """
    def __init__(self, execution_context, top_count):
        super(_QueryExecutionTopEndpointComponent, self).__init__(execution_context)
        self._top_count = top_count

    def next(self):
        if (self._top_count > 0):
            res = next(self._execution_context)
            self._top_count -= 1
            return res
        raise StopIteration

class _QueryExecutionAggregateEndpointComponent(_QueryExecutionEndpointComponent):
    """Represents an endpoint in handling aggregate query.

    It returns only aggreated values.
    """
    def __init__(self, execution_context, aggregate_operators):
        super(_QueryExecutionAggregateEndpointComponent, self).__init__(execution_context)
        self._local_aggregators = []
        self._results = None
        self._result_index = 0
        for operator in aggregate_operators:
            if operator == 'Average':
                self._local_aggregators.append(_AverageAggregator())
            elif operator == 'Count':
                self._local_aggregators.append(_CountAggregator())
            elif operator == 'Max':
                self._local_aggregators.append(_MaxAggregator())
            elif operator == 'Min':
                self._local_aggregators.append(_MinAggregator())
            elif operator == 'Sum':
                self._local_aggregators.append(_SumAggregator())

    def next(self):
        for res in self._execution_context:
            for item in res:
                for operator in self._local_aggregators:
                    if isinstance(item, dict) and len(item.keys()) > 0:
                        operator.aggregate(item['item'])
                    elif isinstance(item, numbers.Number):
                        operator.aggregate(item)
        if self._results is None:
            self._results = []
            for operator in self._local_aggregators:
                self._results.append(operator.get_result())
        if self._result_index < len(self._results):
            res = self._results[self._result_index]
            self._result_index += 1
            return res
        else:
            raise StopIteration



import unittest
import uuid
import threading
import pytest
from time import sleep

from azure.cosmos.http_constants import ResourceType
import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.documents as documents
from azure.cosmos.request_object import _RequestObject
from azure.cosmos.location_cache import LocationCache
from azure.cosmos.global_endpoint_manager import _GlobalEndpointManager
import azure.cosmos.errors as errors
from azure.cosmos.http_constants import StatusCodes, SubStatusCodes, HttpHeaders
import azure.cosmos.retry_utility as retry_utility
import six

class RefreshThread(threading.Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs=None, verbose=None):
        if six.PY2:
            super(RefreshThread, self).__init__(group=group, target=target,
                                      name=name, verbose=verbose)
        else:
            super().__init__()
        self.endpoint_manager = kwargs['endpoint_manager']
    def run(self):
        self.endpoint_manager.force_refresh(None)

@pytest.mark.usefixtures("teardown")
class LocationCacheTest(unittest.TestCase):

    DEFAULT_ENDPOINT = "https://default.documents.azure.com"
    LOCATION_1_ENDPOINT = "https://location1.documents.azure.com"
    LOCATION_2_ENDPOINT = "https://location2.documents.azure.com"
    LOCATION_3_ENDPOINT = "https://location3.documents.azure.com"
    LOCATION_4_ENDPOINT = "https://location4.documents.azure.com"
    REFRESH_TIME_INTERVAL_IN_MS = 1000
    endpoint_by_location = {"location1": LOCATION_1_ENDPOINT,
                            "location2": LOCATION_2_ENDPOINT,
                            "location3": LOCATION_3_ENDPOINT,
                            "location4": LOCATION_4_ENDPOINT}

    def mock_create_db_with_flag_enabled(self, url_connection = None):
        self.database_account = self.create_database_account(True)
        return self.database_account

    def mock_create_db_with_flag_disabled(self, url_connection = None):
        self.database_account = self.create_database_account(False)
        return self.database_account

    def create_spy_client(self, use_multiple_write_locations, enable_endpoint_discovery, is_preferred_locations_list_empty):
        self.preferred_locations = ["location1", "location2", "location3", "location4"]
        connectionPolicy = documents.ConnectionPolicy()
        connectionPolicy.DisableSSLVerification = True
        connectionPolicy.PreferredLocations = [] if is_preferred_locations_list_empty else self.preferred_locations
        connectionPolicy.EnableEndpointDiscovery = enable_endpoint_discovery
        connectionPolicy.UseMultipleWriteLocations = use_multiple_write_locations

        client = cosmos_client.CosmosClient(self.DEFAULT_ENDPOINT, {'masterKey': "SomeKeyValue"}, connectionPolicy)
        return client

    def test_validate_retry_on_session_not_availabe_with_disable_multiple_write_locations_and_endpoint_discovery_disabled(self):
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(False, False, False)
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(False, False, True)
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(False, True, False)
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(False, True, True)
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(True, False, False)
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(True, False, True)
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(True, True, False)
        self.validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(True, True, True)

    def validate_retry_on_session_not_availabe_with_endpoint_discovery_disabled(self, is_preferred_locations_list_empty, use_multiple_write_locations, is_read_request):
        self.counter = 0
        self.OriginalExecuteFunction = retry_utility._ExecuteFunction
        retry_utility._ExecuteFunction = self._MockExecuteFunctionSessionReadFailureOnce
        self.original_get_database_account = cosmos_client.CosmosClient.GetDatabaseAccount
        cosmos_client.CosmosClient.GetDatabaseAccount = self.mock_create_db_with_flag_enabled if use_multiple_write_locations else self.mock_create_db_with_flag_disabled
        enable_endpoint_discovery = False
        client = self.create_spy_client(use_multiple_write_locations, enable_endpoint_discovery, is_preferred_locations_list_empty)
        
        try:
            if is_read_request:
                client.ReadItem("dbs/mydb/colls/mycoll/docs/1")
            else:
                client.CreateItem("dbs/mydb/colls/mycoll/", {'id':'1'})
            self.fail()
        except errors.HTTPFailure as e:
            # not retried
            self.assertEqual(self.counter, 1)
            self.counter = 0
            self.assertEqual(e.status_code, StatusCodes.NOT_FOUND)
            self.assertEqual(e.sub_status, SubStatusCodes.READ_SESSION_NOTAVAILABLE)

        cosmos_client.CosmosClient.GetDatabaseAccount = self.original_get_database_account
        retry_utility._ExecuteFunction = self.OriginalExecuteFunction

    def _MockExecuteFunctionSessionReadFailureOnce(self, function, *args, **kwargs):
        self.counter += 1
        raise errors.HTTPFailure(StatusCodes.NOT_FOUND, "Read Session not available", {HttpHeaders.SubStatus: SubStatusCodes.READ_SESSION_NOTAVAILABLE})

    def test_validate_retry_on_session_not_availabe_with_endpoint_discovery_enabled(self):
        # sequence of chosen endpoints: 
        #     1. Single region, No Preferred Location: 
        #        location1 (default) -> location1 (no preferred location, hence default)
        #     2. Single Region, Preferred Locations present:
        #        location1 (1st preferred location) -> location1 (1st location in DBA's WriteLocation)
        #     3. MultiRegion, Preferred Regions present:
        #        location1 (1st preferred location Read Location) -> location1 (1st location in DBA's WriteLocation) ->
        #        location2 (2nd preferred location Read Location)-> location4 (3rd preferred location Read Location)
        #self.validate_retry_on_session_not_availabe(True, False)
        #self.validate_retry_on_session_not_availabe(False, False)
        self.validate_retry_on_session_not_availabe(False, True)

    def validate_retry_on_session_not_availabe(self, is_preferred_locations_list_empty, use_multiple_write_locations):
        self.counter = 0
        self.OriginalExecuteFunction = retry_utility._ExecuteFunction
        retry_utility._ExecuteFunction = self._MockExecuteFunctionSessionReadFailureTwice
        self.original_get_database_account = cosmos_client.CosmosClient.GetDatabaseAccount
        cosmos_client.CosmosClient.GetDatabaseAccount = self.mock_create_db_with_flag_enabled if use_multiple_write_locations else self.mock_create_db_with_flag_disabled

        enable_endpoint_discovery = True
        self.is_preferred_locations_list_empty = is_preferred_locations_list_empty
        self.use_multiple_write_locations = use_multiple_write_locations
        client = self.create_spy_client(use_multiple_write_locations, enable_endpoint_discovery, is_preferred_locations_list_empty)

        try:
            client.ReadItem("dbs/mydb/colls/mycoll/docs/1")
        except errors.HTTPFailure as e:
            # not retried
            self.assertEqual(self.counter, 4 if use_multiple_write_locations else 2)
            self.counter = 0
            self.assertEqual(e.status_code, StatusCodes.NOT_FOUND)
            self.assertEqual(e.sub_status, SubStatusCodes.READ_SESSION_NOTAVAILABLE)

        cosmos_client.CosmosClient.GetDatabaseAccount = self.original_get_database_account
        retry_utility._ExecuteFunction = self.OriginalExecuteFunction

    def _MockExecuteFunctionSessionReadFailureTwice(self, function, *args, **kwargs):
        request = args[1]
        if self.counter == 0:
            if not self.use_multiple_write_locations:
                expected_endpoint = self.database_account.WritableLocations[0]['databaseAccountEndpoint'] if self.is_preferred_locations_list_empty else self.preferred_locations[0]
            else:
                expected_endpoint = self.endpoint_by_location[self.preferred_locations[0]]
            self.assertFalse(request.should_clear_session_token_on_session_read_failure)
        elif self.counter == 1:
            expected_endpoint = self.database_account.WritableLocations[0]['databaseAccountEndpoint']
            if not self.use_multiple_write_locations:
                self.assertTrue(request.should_clear_session_token_on_session_read_failure)
            else:
                self.assertFalse(request.should_clear_session_token_on_session_read_failure)
        elif self.counter == 2:
            expected_endpoint = self.endpoint_by_location[self.preferred_locations[1]]
            self.assertFalse(request.should_clear_session_token_on_session_read_failure)
        elif self.counter == 3:
            expected_endpoint = self.database_account.ReadableLocations[2]['databaseAccountEndpoint']
            self.assertTrue(request.should_clear_session_token_on_session_read_failure)
        self.assertEqual(expected_endpoint, request.location_endpoint_to_route)
        self.counter += 1
        raise errors.HTTPFailure(StatusCodes.NOT_FOUND, "Read Session not available", {HttpHeaders.SubStatus: SubStatusCodes.READ_SESSION_NOTAVAILABLE})

    def test_validate_location_cache(self):
        self.original_get_database_account = cosmos_client.CosmosClient.GetDatabaseAccount
        cosmos_client.CosmosClient.GetDatabaseAccount = self.mock_get_database_account
        self.get_database_account_hit_counter = 0
        for i in range (0,8):
            use_multiple_write_locations = (i & 1) > 0
            endpoint_discovery_enabled = (i & 2) > 0
            is_preferred_list_empty = (i & 4) > 0
            self.validate_location_cache(use_multiple_write_locations, endpoint_discovery_enabled, is_preferred_list_empty)
        cosmos_client.CosmosClient.GetDatabaseAccount = self.original_get_database_account

    def test_validate_write_endpoint_order_with_client_side_disable_multiple_write_location(self):
        self.original_get_database_account = cosmos_client.CosmosClient.GetDatabaseAccount
        cosmos_client.CosmosClient.GetDatabaseAccount = self.mock_get_database_account
        self.get_database_account_hit_counter = 0
        self.initialize(False, True, False)
        self.assertEqual(self.location_cache.get_write_endpoints()[0], self.LOCATION_1_ENDPOINT)
        self.assertEqual(self.location_cache.get_write_endpoints()[1], self.LOCATION_2_ENDPOINT)
        self.assertEqual(self.location_cache.get_write_endpoints()[2], self.LOCATION_3_ENDPOINT)
        cosmos_client.CosmosClient.GetDatabaseAccount = self.original_get_database_account

    def mock_get_database_account(self, url_connection = None):
        self.get_database_account_hit_counter += 1
        return self.create_database_account(True)

    def create_database_account(self, use_multiple_write_locations):
        database_account = documents.DatabaseAccount()
        database_account._EnableMultipleWritableLocations = use_multiple_write_locations
        database_account._WritableLocations = [
                    {'name': 'location1', 'databaseAccountEndpoint': self.LOCATION_1_ENDPOINT},
                    {'name': 'location2', 'databaseAccountEndpoint': self.LOCATION_2_ENDPOINT},
                    {'name': 'location3', 'databaseAccountEndpoint': self.LOCATION_3_ENDPOINT}
                    ]
        database_account._ReadableLocations = [
                    {'name': 'location1', 'databaseAccountEndpoint': self.LOCATION_1_ENDPOINT},
                    {'name': 'location2', 'databaseAccountEndpoint': self.LOCATION_2_ENDPOINT},
                    {'name': 'location4', 'databaseAccountEndpoint': self.LOCATION_4_ENDPOINT}
                    ]
        return database_account

    def initialize(self, use_multiple_write_locations, enable_endpoint_discovery, is_preferred_locations_list_empty):
        self.database_account = self.create_database_account(use_multiple_write_locations)
        preferred_locations = ["location1", "location2", "location3"]
        self.preferred_locations = [] if is_preferred_locations_list_empty else preferred_locations
        self.location_cache = LocationCache(
                self.preferred_locations,
                self.DEFAULT_ENDPOINT,
                enable_endpoint_discovery,
                use_multiple_write_locations,
                self.REFRESH_TIME_INTERVAL_IN_MS)
        self.location_cache.perform_on_database_account_read(self.database_account)
        connectionPolicy = documents.ConnectionPolicy()
        connectionPolicy.PreferredLocations = self.preferred_locations
        client = cosmos_client.CosmosClient("", {}, connectionPolicy)
        self.global_endpoint_manager = client._global_endpoint_manager

    def validate_location_cache(self, use_multiple_write_locations, endpoint_discovery_enabled, is_preferred_list_empty):
        for write_location_index in range(0,3):
            for read_location_index in range(0,2):
                self.initialize(use_multiple_write_locations, endpoint_discovery_enabled, is_preferred_list_empty)

                current_write_endpoints = self.location_cache.get_write_endpoints()
                current_read_endpoints = self.location_cache.get_read_endpoints()

                for i in range(0, read_location_index):
                    self.location_cache.mark_endpoint_unavailable_for_read(self.database_account.ReadableLocations[i]['databaseAccountEndpoint'])
                    self.global_endpoint_manager.mark_endpoint_unavailable_for_read(self.database_account.ReadableLocations[i]['databaseAccountEndpoint'])

                for i in range(0, write_location_index):
                    self.location_cache.mark_endpoint_unavailable_for_write(self.database_account.WritableLocations[i]['databaseAccountEndpoint'])
                    self.global_endpoint_manager.mark_endpoint_unavailable_for_write(self.database_account.WritableLocations[i]['databaseAccountEndpoint'])

                write_endpoint_by_location = {}
                for dba_location in self.database_account._WritableLocations:
                    write_endpoint_by_location[dba_location['name']] = dba_location['databaseAccountEndpoint']

                read_endpoint_by_location = {}
                for dba_location in self.database_account._ReadableLocations:
                    read_endpoint_by_location[dba_location['name']] = dba_location['databaseAccountEndpoint']

                available_write_endpoints = []
                for i in range(write_location_index, len(self.preferred_locations)):
                    location = self.preferred_locations[i]
                    endpoint = write_endpoint_by_location[location] if location in write_endpoint_by_location else None
                    if endpoint:
                        available_write_endpoints.append(endpoint)

                available_read_endpoints = []
                for i in range(read_location_index, len(self.preferred_locations)):
                    location = self.preferred_locations[i]
                    endpoint = read_endpoint_by_location[location] if location in read_endpoint_by_location else None
                    if endpoint:
                        available_read_endpoints.append(endpoint)

                self.validate_endpoint_refresh(use_multiple_write_locations, endpoint_discovery_enabled, available_write_endpoints, available_read_endpoints, write_location_index > 0)
                self.validate_global_endpoint_location_cache_refresh()
                self.validate_request_endpoint_resolution(use_multiple_write_locations, endpoint_discovery_enabled, available_write_endpoints, available_read_endpoints)

                # wait for TTL on unavailablity info
                sleep(1.5)

                self.assertEquals(current_write_endpoints, self.location_cache.get_write_endpoints())
                self.assertEquals(current_read_endpoints, self.location_cache.get_read_endpoints())

    def validate_global_endpoint_location_cache_refresh(self):
        self.get_database_account_hit_counter = 0
        refresh_threads = []
        for i in range(10):
            refresh_thread = RefreshThread(kwargs={'endpoint_manager':self.global_endpoint_manager})
            refresh_thread.start()
            refresh_threads.append(refresh_thread)

        for i in range(10):
            refresh_threads[i].join()

        self.assertTrue(self.get_database_account_hit_counter <= 1)

        for i in range(10):
            refresh_thread = RefreshThread(kwargs={'endpoint_manager':self.global_endpoint_manager})
            refresh_thread.start()
            refresh_thread.join()

        self.assertTrue(self.get_database_account_hit_counter <= 1)

    def validate_endpoint_refresh(self, use_multiple_write_locations, endpoint_discovery_enabled, preferred_available_write_endpoints,
                                  preferred_available_read_endpoints, is_first_write_endpoint_unavailable):
        should_refresh_endpoints = self.location_cache.should_refresh_endpoints()

        is_most_preferred_location_unavailable_for_read = False
        is_most_preferred_location_unavailable_for_write = False if use_multiple_write_locations else is_first_write_endpoint_unavailable

        if (len(self.preferred_locations) > 0):
            most_preferred_read_location_name = None
            for preferred_location in self.preferred_locations:
                for read_location in self.database_account._ReadableLocations:
                    if read_location['name'] == preferred_location:
                        most_preferred_read_location_name = preferred_location
                        break
                if most_preferred_read_location_name:
                    break

            most_preferred_read_endpoint = self.endpoint_by_location[most_preferred_read_location_name]
            is_most_preferred_location_unavailable_for_read = True if len(preferred_available_read_endpoints) == 0 else preferred_available_read_endpoints[0] != most_preferred_read_endpoint

            most_preferred_write_location_name = None
            for preferred_location in self.preferred_locations:
                for write_location in self.database_account._WritableLocations:
                    if write_location['name'] == preferred_location:
                        most_preferred_write_location_name = preferred_location
                        break
                if most_preferred_write_location_name:
                    break

            most_preferred_write_endpoint = self.endpoint_by_location[most_preferred_write_location_name]

            if use_multiple_write_locations:
                is_most_preferred_location_unavailable_for_write = True if len(preferred_available_write_endpoints) == 0 else preferred_available_write_endpoints[0] != most_preferred_write_endpoint

        if not endpoint_discovery_enabled:
            self.assertFalse(should_refresh_endpoints)
        else:
            self.assertEquals(is_most_preferred_location_unavailable_for_read or is_most_preferred_location_unavailable_for_write, should_refresh_endpoints)

    def validate_request_endpoint_resolution(self, use_multiple_write_locations, endpoint_discovery_enabled,
                                             available_write_endpoints, available_read_endpoints):
        write_locations = self.database_account._WritableLocations
        if not endpoint_discovery_enabled:
            first_available_write_endpoint = self.DEFAULT_ENDPOINT
            second_available_write_endpoint = self.DEFAULT_ENDPOINT
        elif not use_multiple_write_locations:
            first_available_write_endpoint = write_locations[0]['databaseAccountEndpoint']
            second_available_write_endpoint = write_locations[1]['databaseAccountEndpoint']
        elif len(available_write_endpoints) > 1:
            first_available_write_endpoint = available_write_endpoints[0]
            second_available_write_endpoint = available_write_endpoints[1]
        elif len(available_write_endpoints) > 0:
            first_available_write_endpoint = available_write_endpoints[0]
            write_endpoint = write_locations[0]['databaseAccountEndpoint']
            second_available_write_endpoint = write_endpoint if write_endpoint != first_available_write_endpoint else available_write_endpoints[1]
        else:
            first_available_write_endpoint = self.DEFAULT_ENDPOINT
            second_available_write_endpoint = self.DEFAULT_ENDPOINT

        if not endpoint_discovery_enabled:
            first_available_read_endpoint = self.DEFAULT_ENDPOINT
        elif len(self.preferred_locations) == 0:
            first_available_read_endpoint = first_available_write_endpoint
        elif len(available_read_endpoints) > 0:
            first_available_read_endpoint = available_read_endpoints[0]
        else:
            first_available_read_endpoint = self.endpoint_by_location[self.preferred_locations[0]]

        first_write_enpoint = self.DEFAULT_ENDPOINT if not endpoint_discovery_enabled else self.database_account.WritableLocations[0]['databaseAccountEndpoint']

        second_write_enpoint = self.DEFAULT_ENDPOINT if not endpoint_discovery_enabled else self.database_account.WritableLocations[1]['databaseAccountEndpoint']

        # If current write endpoint is unavailable, write endpoints order doesn't change
        # All write requests flip-flop between current write and alternate write endpoint
        write_endpoints = self.location_cache.get_write_endpoints()
        self.assertTrue(first_available_write_endpoint == write_endpoints[0])
        self.assertTrue(second_available_write_endpoint == self.resolve_endpoint_for_write_request(ResourceType.Document, True))
        self.assertTrue(first_available_write_endpoint == self.resolve_endpoint_for_write_request(ResourceType.Document, False))
        
        # Writes to other resource types should be directed to first/second write endpoint
        self.assertTrue(first_write_enpoint == self.resolve_endpoint_for_write_request(ResourceType.Database, False))
        self.assertTrue(second_write_enpoint == self.resolve_endpoint_for_write_request(ResourceType.Database, True))

        # Reads should be directed to available read endpoints regardless of resource type
        self.assertTrue(first_available_read_endpoint == self.resolve_endpoint_for_read_request(True))
        self.assertTrue(first_available_read_endpoint == self.resolve_endpoint_for_read_request(False))

    def resolve_endpoint_for_read_request(self, master_resource_type):
        operation_type = documents._OperationType.Read
        resource_type = ResourceType.Database if master_resource_type else ResourceType.Document
        request = _RequestObject(resource_type, operation_type)
        return self.location_cache.resolve_service_endpoint(request)

    def resolve_endpoint_for_write_request(self, resource_type, use_alternate_write_endpoint):
        operation_type = documents._OperationType.Create
        request = _RequestObject(resource_type, operation_type)
        request.route_to_location_with_preferred_location_flag(1 if use_alternate_write_endpoint else 0, ResourceType.IsCollectionChild(resource_type))
        return self.location_cache.resolve_service_endpoint(request)



#The MIT License (MIT)
#Copyright (c) 2014 Microsoft Corporation

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import unittest
import pytest
import azure.cosmos.documents as documents
import azure.cosmos.cosmos_client as cosmos_client
from azure.cosmos.routing.routing_map_provider import _PartitionKeyRangeCache
from azure.cosmos.routing import routing_range as routing_range
import test.test_config as test_config

#IMPORTANT NOTES:
  
#      Most test cases in this file create collections in your Azure Cosmos account.
#      Collections are billing entities.  By running these test cases, you may incur monetary costs on your account.
  
#      To Run the test, replace the two member fields (masterKey and host) with values 
#   associated with your Azure Cosmos account.

@pytest.mark.usefixtures("teardown")
class RoutingMapEndToEndTests(unittest.TestCase):
    """Routing Map Functionalities end to end Tests.
    """

    host = test_config._test_config.host
    masterKey = test_config._test_config.masterKey
    connectionPolicy = test_config._test_config.connectionPolicy
    client = cosmos_client.CosmosClient(host, {'masterKey': masterKey}, connectionPolicy)
    collection_link = test_config._test_config.create_multi_partition_collection_with_custom_pk_if_not_exist(client)['_self']

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")

    def test_read_partition_key_ranges(self):
        partition_key_ranges = list(self.client._ReadPartitionKeyRanges(self.collection_link))
        #"the number of expected partition ranges returned from the emulator is 5."
        self.assertEqual(5, len(partition_key_ranges))
        
    def test_routing_map_provider(self):
        partition_key_ranges = list(self.client._ReadPartitionKeyRanges(self.collection_link))

        routing_mp = _PartitionKeyRangeCache(self.client)
        overlapping_partition_key_ranges = routing_mp.get_overlapping_ranges(self.collection_link, routing_range._Range("", "FF", True, False))
        self.assertEqual(len(overlapping_partition_key_ranges), len(partition_key_ranges))
        self.assertEqual(overlapping_partition_key_ranges, partition_key_ranges)

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()